/**
 * Wrapper around Tango thread
 */
module std.thread;
import std.compat;

private import TT = tango.core.Thread;


/**
 * Thrown for errors.
 */
class ThreadError : Error
{
    this(char[] s)
    {
        super("Thread error: " ~ s);
    }
}


/**
 * One of these is created for each thread.
 */
class Thread
{
    /**
     * Constructor used by classes derived from Thread that override run().
     */
    this(size_t stacksize = 0)
    {
        t = new TT.Thread(&runImpl, stacksize);
    }

    /**
     * Constructor used by classes derived from Thread that override run().
     */
    this(int (*fp)(void *), void *arg, size_t stacksize = 0)
    {
        this.fp = fp;
        this.arg = arg;
        this(stacksize);
    }

    /**
     * Constructor used by classes derived from Thread that override run().
     */
    this(int delegate() dg, size_t stacksize = 0)
    {
        this.dg = dg;
        this(stacksize);
    }

    /**
     * Create a new thread and start it running. The new thread initializes
     * itself and then calls run(). start() can only be called once.
     */
    void start()
    {
        t.start();
    }

    /**
     * Entry point for a thread. If not overridden, it calls the function
     * pointer fp and argument arg passed in the constructor, or the delegate
     * dg.
     * Returns: the thread exit code, which is normally 0.
     */
    int run()
    {
        if (fp)
            return fp(arg);
        else if (dg)
            return dg();
        assert(0);
    }

    /*****************************
     * Wait for this thread to terminate.
     * Simply returns if thread has already terminated.
     * Throws: $(B ThreadError) if the thread hasn't begun yet or
     * is called on itself.
     */
    void wait()
    {
        t.join();
    }

    /******************************
     * Wait for this thread to terminate or until milliseconds time has
     * elapsed, whichever occurs first.
     * Simply returns if thread has already terminated.
     * Throws: $(B ThreadError) if the thread hasn't begun yet or
     * is called on itself.
     */
    void wait(uint milliseconds)
    {
        t.join();
    }

    /**
     * The state of a thread.
     */
    enum TS
    {
        INITIAL,        /// The thread hasn't been started yet.
        RUNNING,        /// The thread is running or paused.
        TERMINATED,     /// The thread has ended.
        FINISHED        /// The thread has been cleaned up
    }

    /**
     * Returns the state of a thread.
     */
    TS getState()
    {
	return t.isRunning ? TS.RUNNING : TS.TERMINATED;
    }

    /**
     * The priority of a thread.
     */
    enum PRIORITY
    {
        INCREASE,       /// Increase thread priority
        DECREASE,       /// Decrease thread priority
        IDLE,           /// Assign thread low priority
        CRITICAL        /// Assign thread high priority
    }

    /**
     * Adjust the priority of this thread.
     * Throws: ThreadError if cannot set priority
     */
    void setPriority(PRIORITY p)
    {
        error("not supported in Tango thread proxy");
    }

    /**
     * Returns true if this thread is the current thread.
     */
    bool isSelf()
    {
        error("not supported in Tango thread proxy");
        return false;
    }

    /**
     * Returns a reference to the Thread for the thread that called the
     * function.
     */
    static Thread getThis()
    {
        error("not supported in Tango thread proxy");
        return null;
    }

    /**
     * Returns an array of all the threads currently running.
     */
    static Thread[] getAll()
    {
        error("not supported in Tango thread proxy");
        return null;
    }

    /**
     * Suspend execution of this thread.
     */
    void pause()
    {
        error("not supported in Tango proxy class");
    }

    /**
     * Resume execution of this thread.
     */
    void resume()
    {
        error("not supported in Tango proxy class");
    }

    /**
     * Suspend execution of all threads but this thread.
     */
    static void pauseAll()
    {
        TT.thread_suspendAll();
    }

    /**
     * Resume execution of all paused threads.
     */
    static void resumeAll()
    {
        TT.thread_resumeAll();
    }

    /**
     * Give up the remainder of this thread's time slice.
     */
    static void yield()
    {
        t.yield();
    }

    /**
     *
     */
    static uint nthreads = 1;

private:
    TS state;

    int (*fp)(void *);
    void *arg;

    int delegate() dg;

    static void error(char[] msg)
    {
        throw new ThreadError(msg);
    }

    void runImpl()
    {
        run();
    }

    private TT.Thread t;
}
